#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from enum import StrEnum
from typing import Annotated

from fastapi import Body
from pydantic import BaseModel, Field

from cml_mcp.schemas.annotations import AnnotationResponse, AnnotationUpdate
from cml_mcp.schemas.common import MACAddress, UUID4Type
from cml_mcp.schemas.interfaces import InterfaceLabel, InterfaceSlot, InterfaceType
from cml_mcp.schemas.labs import LabDescription, LabNotes, LabTitle
from cml_mcp.schemas.links import (
    LinkConditionConfiguration,
    LinkLabel,
    LinkWithConditionConfig,
)
from cml_mcp.schemas.nodes import NodeConfiguration, NodeDefined
from cml_mcp.schemas.smart_annotations import (
    SmartAnnotationBase,
    SmartAnnotationResponse,
)

TopologyID = Annotated[str, Field(min_length=1, max_length=64, description="Element ID.", examples=["l1"])]


class TopologySchemaVersionEnum(StrEnum):
    v0_0_1 = "0.0.1"
    v0_0_2 = "0.0.2"
    v0_0_3 = "0.0.3"
    v0_0_4 = "0.0.4"
    v0_0_5 = "0.0.5"
    v0_1_0 = "0.1.0"
    v0_2_0 = "0.2.0"
    v0_2_1 = "0.2.1"
    v0_2_2 = "0.2.2"
    v0_3_0 = "0.3.0"


TopologySchemaVersion = Annotated[
    TopologySchemaVersionEnum,
    Field(description="Topology schema version.", examples=["0.2.2"]),
]


class LabTopology(BaseModel, extra="forbid"):
    version: TopologySchemaVersion = Field(...)
    title: LabTitle = Field(default=None)
    description: LabDescription = Field(default=None)
    notes: LabNotes = Field(default=None)


class LabTopologyWithOwner(LabTopology, extra="forbid"):
    owner: UUID4Type = Field(default=None)


class InterfaceTopology(BaseModel, extra="forbid"):
    id: TopologyID = Field(...)
    type: InterfaceType = Field(...)
    node: TopologyID = Field(default=None)
    label: InterfaceLabel = Field(default=None)
    slot: InterfaceSlot = Field(default=None)
    mac_address: MACAddress = Field(default=None)


class NodeTopology(NodeDefined, extra="forbid"):
    id: TopologyID = Field(...)
    interfaces: list[InterfaceTopology] = Field(...)
    configuration: NodeConfiguration = Field(default=None)


class LinkTopology(BaseModel, extra="forbid"):
    id: TopologyID = Field(...)
    i1: TopologyID = Field(...)
    i2: TopologyID = Field(...)
    n1: TopologyID = Field(...)
    n2: TopologyID = Field(...)
    label: LinkLabel = Field(default=None)
    conditioning: LinkConditionConfiguration = Field(default_factory=dict)


class Topology(BaseModel, extra="forbid"):
    """
    A JSON object that describes topology, including its nodes, interfaces,
    links, and details, such as the user-supplied title and notes.
    """

    nodes: list[NodeTopology] = Field(...)
    links: list[LinkTopology] = Field(...)
    lab: LabTopology = Field(...)
    annotations: list[AnnotationUpdate] = Field(default_factory=list)
    smart_annotations: list[SmartAnnotationBase] = Field(default_factory=list)


ImportTopologyBody = Annotated[
    Topology,
    Body(description="This request body is a JSON object that describes a single topology."),
]


class TopologyResponse(Topology, extra="forbid"):
    """The response body is a JSON topology object."""

    links: list[LinkWithConditionConfig] = Field(...)
    lab: LabTopologyWithOwner = Field(...)
    annotations: list[AnnotationResponse] = Field(default=None)
    smart_annotations: list[SmartAnnotationResponse] = Field(default=None)



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
import re
from enum import StrEnum
from typing import Annotated, Literal

from pydantic import BaseModel, Field, conlist, model_validator

from cml_mcp.schemas.common import DefinitionID, FilePath
from cml_mcp.schemas.nodes import NodeConfigurationContent, NodeParameters


class DriverTypes(StrEnum):
    ASAV = "asav"
    ALPINE = "alpine"
    CAT9K = "cat9k"
    COREOS = "coreos"
    CSR1000V = "csr1000v"
    EXTERNAL_CONNECTOR = "external_connector"
    IOL = "iol"
    IOL_L2 = "iol-l2"
    IOSV = "iosv"
    IOSVL2 = "iosvl2"
    IOSXRV = "iosxrv"
    IOSXRV9000 = "iosxrv9000"
    LXC = "lxc"
    NXOSV = "nxosv"
    NXOSV9000 = "nxosv9000"
    PAGENT = "pagent"
    SERVER = "server"
    TREX = "trex"
    UBUNTU = "ubuntu"
    UNMANAGED_SWITCH = "unmanaged_switch"
    WAN_EMULATOR = "wan_emulator"


class NicDrivers(StrEnum):
    VIRTIO = "virtio"
    E1000 = "e1000"
    RTL8139 = "rtl8139"
    VMXNET3 = "vmxnet3"
    E1000E = "e1000e"
    E1000_82544GC = "e1000-82544gc"
    E1000_82545EM = "e1000-82545em"
    I82550 = "i82550"
    I82551 = "i82551"
    I82557A = "i82557a"
    I82557B = "i82557b"
    I82557C = "i82557c"
    I82558A = "i82558a"
    I82558B = "i82558b"
    I82559A = "i82559a"
    I82559B = "i82559b"
    I82559C = "i82559c"
    I82559ER = "i82559er"
    I82562 = "i82562"
    I82801 = "i82801"


class LibvirtDomainDrivers(StrEnum):
    DOCKER = "docker"
    IOL = "iol"
    KVM = "kvm"
    LXC = "lxc"
    NONE = "none"


class DiskDrivers(StrEnum):
    IDE = "ide"
    SATA = "sata"
    VIRTIO = "virtio"


class VideoModels(StrEnum):
    VGA = "vga"
    CIRRUS = "cirrus"
    VMVGA = "vmvga"
    QXL = "qxl"
    XEN = "xen"
    VIRTIO = "virtio"
    NONE = "none"


class DeviceNature(StrEnum):
    ROUTER = "router"
    SWITCH = "switch"
    SERVER = "server"
    HOST = "host"
    CLOUD = "cloud"
    FIREWALL = "firewall"
    EXTERNAL_CONNECTOR = "external_connector"


class VideoDevice(BaseModel, extra="forbid"):
    memory: int = Field(..., description="Video Memory.", ge=1, le=128)
    model: VideoModels = Field(default=None, description="Video Model.")


class LinuxNative(BaseModel, extra="forbid"):
    """Base for simulation objects."""

    libvirt_domain_driver: LibvirtDomainDrivers = Field(..., description="Domain Driver.")
    driver: DriverTypes = Field(..., description="Simulation Driver.")
    disk_driver: DiskDrivers = Field(default=None, description="Disk Driver.")
    efi_boot: bool = Field(default=None, description="If set, use EFI boot for the VM.")
    efi_code: FilePath = Field(default=None, description="EFI code file path; if unset, use default.")
    efi_vars: FilePath = Field(
        default=None,
        description="EFI NVRAM var template path; if unset, the code file "
        "is made writable; if set to constant 'stateless', "
        "the code file is marked stateless.",
    )
    machine_type: str = Field(
        default=None,
        description="QEMU machine type, defaults to pc; q35 is more modern.",
        min_length=1,
        max_length=32,
    )
    ram: int | None = Field(default=None, description="Memory in MiB.", ge=1, le=1048576)
    cpus: int | None = Field(default=None, description="CPUs.", ge=1, le=128)
    cpu_limit: int | None = Field(default=None, description="CPU Limit.", ge=20, le=100)
    cpu_model: str | None = Field(
        default=None,
        min_length=1,
        max_length=64,
        pattern=re.compile(r"^[a-zA-Z\d-]{1,32}(,[+!?^-][a-z\d._]{1,16})*(?![\n\r])$"),
    )
    nic_driver: NicDrivers = Field(
        default=None,
        description="Network Driver.",
    )
    data_volume: int = Field(default=None, description="Data Disk Size in GiB.", ge=0, le=4096)
    boot_disk_size: int = Field(default=None, description="Boot Disk Size in GiB.", ge=0, le=4096)
    video: VideoDevice = Field(default=None, description="If present, then VNC can be used with the node VM.")
    enable_rng: bool = Field(
        default=True,
        description="If set, use a random number generator.",
    )
    enable_tpm: bool = Field(
        default=False,
        description="If set, enable an emulated TPM 2.0.",
    )

    @model_validator(mode="after")
    def validate(self):
        if self.libvirt_domain_driver == LibvirtDomainDrivers.KVM:
            required = [self.cpus, self.ram, self.nic_driver, self.disk_driver]
            if not all(required):
                raise ValueError(f"{required} must be specified when libvirt_domain_driver=kvm")

        if self.ram is None and self.libvirt_domain_driver in [
            LibvirtDomainDrivers.DOCKER,
            LibvirtDomainDrivers.LXC,
            LibvirtDomainDrivers.IOL,
        ]:
            raise ValueError("ram must be specified when libvirt_domain_driver in [docker, lxc, iol]")
        return self


PhysicalField = Annotated[str, Field(min_length=1, max_length=32)]

LoopBackField = Annotated[str, Field(min_length=1, max_length=32)]

ManagementField = Annotated[str, Field(min_length=1, max_length=32)]


class Interfaces(BaseModel, extra="forbid"):
    """
    Interface configurations.
    """

    serial_ports: int = Field(..., description="Number of serial Ports (console, aux, ...).", ge=0, le=4)
    default_console: int | None = Field(
        default=None,
        description="Default serial port for console connections.",
        ge=0,
        le=4,
    )
    physical: conlist(PhysicalField, min_length=1) = Field(..., description="List of physical interfaces.")
    has_loopback_zero: bool = Field(..., description="Has `loopback0` interface (used with ANK).")
    min_count: int = Field(
        default=None,
        description="Minimal number of physical interfaces needed to start a node.",
        ge=0,
        le=64,
    )
    default_count: int = Field(default=None, description="Default number of physical interfaces.", ge=1, le=64)
    iol_static_ethernets: Literal[0, 4, 8, 12, 16] = Field(
        default=None,
        description="Only for IOL nodes, the number of static Ethernet interfaces"
        " preceding any serial interface; default 0 means "
        "all interfaces are Ethernet.",
    )
    loopback: conlist(LoopBackField, min_length=1) = Field(default=None, description="List of loopback interfaces.")
    management: conlist(ManagementField, min_length=1) = Field(default=None, description="List of management interfaces.")

    @model_validator(mode="after")
    def validate_one_of(self):
        if self.has_loopback_zero and not self.loopback:
            raise ValueError("loopback must be specified when has_loopback_zero=True")
        return self


class VMProperties(BaseModel, extra="forbid"):
    """Virtual Machine properties."""

    ram: bool = Field(..., description="RAM")
    cpus: bool = Field(..., description="CPU Count.")
    data_volume: bool = Field(..., description="Data Disk Size.")
    boot_disk_size: bool = Field(..., description="Boot Disk Size.")
    cpu_limit: bool | None = Field(default=None, description="CPU Limit.")


CompletedNode = Annotated[str, Field(max_length=128)]


class Boot(BaseModel, extra="forbid"):
    # just a random upper bound
    timeout: int = Field(..., description="Timeout (seconds).", examples=[60], le=86400)
    completed: conlist(CompletedNode, min_length=1) = Field(
        default=None,
        description='A list of strings which should be matched to determine when the node is "ready".',
        examples=[[CompletedNode("string")]],
    )
    uses_regex: bool = Field(
        default=None,
        description="Whether the strings in `completed` should be treated as regular expressions or not.",
    )


class UsageEstimations(BaseModel, extra="forbid"):
    """
    Estimated resource usage.
    """

    cpus: int | None = Field(
        default=None,
        description=("Estimated CPUs usage in one-hundred-part shares of whole CPUs " "(up to 128 CPUs / 12800 shares)."),
        ge=1,
        le=12800,
        examples=[40],
    )
    ram: int | None = Field(
        default=None,
        description="Estimated RAM usage in MiB.",
        ge=1,
        le=1048576,
        examples=[50],
    )
    disk: int | None = Field(
        default=None,
        description="Estimated Disk usage in MiB.",
        ge=1,
        le=4194304,
        examples=[500],
    )


class Simulation(BaseModel, extra="forbid"):
    """
    Simulation parameters.
    """

    linux_native: LinuxNative = Field(..., description="Linux native simulation configuration.")
    parameters: NodeParameters = Field(default=None, description="Node-specific parameters.")
    usage_estimations: UsageEstimations = Field(default=None, description="Estimated resource usage parameters.")


class General(BaseModel, extra="forbid"):
    """
    General information for the node type.
    """

    nature: DeviceNature = Field(..., description='The "nature" / kind of the node type defined here.')
    description: str = Field(default=None, description="A description of the node type.", max_length=4096)
    read_only: bool = Field(
        default=False,
        description="Whether the node definition can be updated and deleted.",
    )


class ConfigurationDrivers(StrEnum):
    asav = "asav"
    alpine = "alpine"
    cat9000v = "cat9000v"
    coreos = "coreos"
    csr1000v = "csr1000v"
    desktop = "desktop"
    fmcv = "fmcv"
    ftdv = "ftdv"
    iosv = "iosv"
    iosvl2 = "iosvl2"
    iosxrv = "iosxrv"
    iosxrv9000 = "iosxrv9000"
    lxc = "lxc"
    nxosv = "nxosv"
    nxosv9000 = "nxosv9000"
    pagent = "pagent"
    sdwan = "sdwan"
    sdwan_edge = "sdwan_edge"
    sdwan_manager = "sdwan_manager"
    server = "server"
    trex = "trex"
    ubuntu = "ubuntu"
    wan_emulator = "wan_emulator"


class ConfigurationMediaTypes(StrEnum):
    iso = "iso"
    fat = "fat"
    raw = "raw"
    ext4 = "ext4"


class ConfigurationGenerator(BaseModel, extra="forbid"):
    """
    Generator configuration details.
    """

    driver: ConfigurationDrivers | None = Field(..., description="Configuration Driver.")


class ConfigurationFile(BaseModel, extra="forbid"):
    """
    Configuration file details.
    """

    editable: bool = Field(..., description="Is the configuration file editable?")
    name: str = Field(
        ...,
        description="The name of the configuration file.",
        min_length=1,
        max_length=64,
    )
    content: NodeConfigurationContent = Field(default=None)


class ConfigurationProvisioning(BaseModel, extra="forbid"):
    """
    Provisioning configuration details.
    """

    files: conlist(ConfigurationFile, min_length=1) = Field(..., description="List of node configuration file objects.")
    media_type: ConfigurationMediaTypes = Field(..., description="The type of the configuration media.")
    volume_name: str = Field(
        ...,
        description="The volume name of the configuration media.",
        min_length=1,
        max_length=32,
    )


class Configuration(BaseModel, extra="forbid"):
    """
    Node definition configuration details.
    """

    generator: ConfigurationGenerator = Field(..., description="Generator configuration details.")
    provisioning: ConfigurationProvisioning = Field(default=None, description="Provisioning configuration details.")


class Device(BaseModel, extra="forbid"):
    interfaces: Interfaces = Field(...)


class Icons(StrEnum):
    router = "router"
    switch = "switch"
    server = "server"
    host = "host"
    cloud = "cloud"
    firewall = "firewall"
    access_point = "access_point"
    wl = "wl"


class Ui(BaseModel, extra="forbid"):
    """
    UI-related settings for the node type.
    """

    label_prefix: str = Field(
        ...,
        description="The textual prefix for node labels.",
        min_length=1,
        max_length=32,
    )
    icon: Icons = Field(..., description="The icon to use with this node type.")
    label: str = Field(..., description="The node type label.", min_length=1, max_length=32)
    visible: bool = Field(..., description="Determines visibility in the UI for this node type.")
    group: Literal["Cisco", "Others"] = Field(default=None, description="Intended to group similar node types (unused).")
    description: str = Field(
        default=None,
        description="The description of the node type (can be Markdown).",
        max_length=4096,
    )
    has_configuration: bool = Field(default=None)
    show_ram: bool = Field(default=None)
    show_cpus: bool = Field(default=None)
    show_cpu_limit: bool = Field(default=None)
    show_data_volume: bool = Field(default=None)
    show_boot_disk_size: bool = Field(default=None)
    has_config_extraction: bool = Field(default=None)


class Inherited(BaseModel, extra="forbid"):
    image: VMProperties = Field(...)
    node: VMProperties = Field(...)


class Pyats(BaseModel, extra="forbid"):
    os: str = Field(
        ...,
        description="The operating system as defined / understood by pyATS.",
        min_length=1,
        max_length=32,
    )
    series: str | None = Field(
        default=None,
        description="The device series as defined by pyATS / Unicon.",
        min_length=1,
        max_length=32,
    )
    model: str = Field(
        default=None,
        description="The device model as defined by pyATS / Unicon.",
        min_length=1,
        max_length=32,
    )
    use_in_testbed: bool = Field(default=None, description="Use this device in an exported testbed?")
    username: str | None = Field(
        default=None,
        description="Use this username with pyATS / Unicon when interacting with this node type.",
        min_length=1,
        max_length=64,
    )
    password: str | None = Field(
        default=None,
        description="Use this password with pyATS / Unicon when interacting with this node type.",
        min_length=1,
        max_length=128,
    )
    config_extract_command: str | None = Field(
        default=None,
        description="This is the CLI command to use when configurations " "should be extracted from a device of this node type.",
        max_length=4096,
    )


Schema = Annotated[
    str,
    Field(
        description="The schema version used for this node type.",
        min_length=1,
        max_length=32,
        examples=["0.0.1"],
    ),
]


class NodeDefinition(BaseModel, extra="forbid"):
    id: DefinitionID = Field(
        ...,
        description="""
            A symbolic name used to identify this node definition, such as `iosv` or
            `asav`.
        """,
    )
    boot: Boot = Field(...)
    sim: Simulation = Field(...)
    general: General = Field(...)
    configuration: Configuration = Field(...)
    device: Device = Field(...)
    ui: Ui = Field(...)
    inherited: Inherited = Field(default=None)
    pyats: Pyats = Field(default=None)
    schema_version: Schema = Field(default=None)


class SimulationUi(BaseModel, extra="forbid"):
    """
    Simplified simulation parameters.
    """

    parameters: NodeParameters = Field(default=None, description="Node-specific parameters.")
    ram: int | None = Field(default=None, ge=0)
    cpus: int | None = Field(default=None, ge=0)
    cpu_limit: int | None = Field(default=100, ge=20, le=100)
    data_volume: int | None = Field(default=None, ge=0)
    boot_disk_size: int | None = Field(default=None, ge=0)
    console: bool | None = Field(default=None)
    simulate: bool | None = Field(default=None)
    custom_mac: bool | None = Field(default=None)
    vnc: bool | None = Field(default=None)


class SimplifiedNodeDefinitionResponse(BaseModel, extra="allow"):
    id: DefinitionID = Field(
        ...,
        description="""
            A symbolic name used to identify this node definition, such as `iosv` or
            `asav`.
        """,
    )
    general: General = Field(...)
    device: Device = Field(...)
    ui: Ui = Field(...)
    sim: SimulationUi = Field(...)
    image_definitions: list[DefinitionID] = Field(default=None)


NodeDefinitionList = Annotated[
    list[NodeDefinition],
    Field(description="Array of Node Definitions"),
]



# Copyright (c) 2025  Joe Clarke <jclarke@cisco.com>
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

import logging
import os
import tempfile

import httpx
from fastmcp import FastMCP, Context
from virl2_client.models.cl_pyats import ClPyats

from cml_mcp.cml_client import CMLClient
from cml_mcp.schemas.common import UserName, UUID4Type
from cml_mcp.schemas.labs import Lab, LabTitle
from cml_mcp.schemas.node_definitions import NodeDefinition, SimplifiedNodeDefinitionResponse
from cml_mcp.schemas.nodes import Node, NodeConfigurationContent, NodeLabel
from cml_mcp.schemas.system import SystemHealth
from cml_mcp.schemas.topologies import Topology
from cml_mcp.settings import settings

# Set up logging
logging.basicConfig(level=logging.DEBUG, format="%(asctime)s %(levelname)s %(threadName)s %(name)s: %(message)s")
logger = logging.getLogger("cml-mcp")


server_mcp = FastMCP(
    "Cisco Modeling Labs MCP Server",
    dependencies=["httpx", "fastmcp", "fastapi", "pydantic_strict_partial", "typer", "virl2_client", "pyats[full]"],
)
cml_client = CMLClient(host=str(settings.cml_url), username=settings.cml_username, password=settings.cml_password)


async def get_all_labs() -> list[UUID4Type]:
    """
    Get all labs from the CML server.

    Returns:
        list[UUID4Type]: A list of lab IDs.
    """
    labs = await cml_client.get("/labs", params={"show_all": True})
    return [UUID4Type(lab) for lab in labs]


@server_mcp.tool
async def get_cml_labs(user: UserName | None = None) -> list[Lab] | dict[str, str]:
    """
    Get the list of labs for a specific user or all labs if the user is an admin.

    Args:
        user (UserName | None): The username to filter labs by. If None, defaults to the configured username.

    Returns:
        list[Lab] | dict[str, str]: A list of Lab objects owned by the user, or a dictionary with an error message if an error occurs.
    """
    if not user:
        user = settings.cml_username  # Default to the configured username

    try:
        # If the requested user is not the configured user and is not an admin, deny access
        if str(user) != settings.cml_username and not await cml_client.is_admin():
            return {"error": "User is not an admin and cannot view all labs."}
        ulabs = []
        # Get all labs from the CML server
        labs = await get_all_labs()
        for lab in labs:
            # For each lab, get its details
            lab_details = await cml_client.get(f"/labs/{lab}")
            # Only include labs owned by the specified user
            if lab_details.get("owner_username") == str(user):
                ulabs.append(Lab(**lab_details))
        return ulabs
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
async def get_cml_status() -> SystemHealth | dict[str, str]:
    """
    Get the status of the CML server.

    Returns:
        SystemHealth: The health/status of the CML server.
        dict[str, str]: An error dictionary if an exception occurs.
    """
    try:
        status = await cml_client.get("/system_health")
        return SystemHealth(**status)
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
async def get_cml_node_definitions() -> list[SimplifiedNodeDefinitionResponse] | dict[str, str]:
    """
    Get the list of node definitions from the CML server.

    Returns:
        list[SimplifiedNodeDefinitionResponse]: A list of simplified node definitions.
        dict[str, str]: An error dictionary if an exception occurs.
    """
    try:
        node_definitions = await cml_client.get("/simplified_node_definitions")
        return [SimplifiedNodeDefinitionResponse(**nd) for nd in node_definitions]
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
async def get_node_definition_detail(nid: UUID4Type) -> NodeDefinition | dict[str, str]:
    """
    Get detailed information about a specific node definition by its ID.

    Args:
        nid (UUID4Type): The node definition ID.

    Returns:
        NodeDefinition: The node definition details.
        dict[str, str]: An error dictionary if an exception occurs.
    """
    try:
        node_definition = await cml_client.get(f"/node_definitions/{nid}")
        return NodeDefinition(**node_definition)
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
async def create_lab_topology(topology: Topology) -> UUID4Type | dict[str, str]:
    """
    Create a new lab topology.

    Args:
        topology (Topology): The topology definition.

    Returns:
        UUID4Type: The new lab topology ID if successful.
        dict[str, str]: An error dictionary if an error occurs.
    """
    try:
        resp = await cml_client.post("/import", data=topology.model_dump())
        return UUID4Type(resp["id"])
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
async def start_cml_lab(lid: UUID4Type) -> None | dict[str, str]:
    """
    Start a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        None: If successful.
        dict[str, str]: An error dictionary if an error occurs.
    """
    try:
        await cml_client.put(f"/labs/{lid}/start")
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


async def stop_lab(lid: UUID4Type) -> None:
    """
    Stop a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        None: If successful.
    """
    await cml_client.put(f"/labs/{lid}/stop")


async def wipe_lab(lid: UUID4Type) -> None:
    """
    Wipe a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        None: If successful.
    """
    await cml_client.put(f"/labs/{lid}/wipe")


@server_mcp.tool
async def stop_cml_lab(lid: UUID4Type) -> None | dict[str, str]:
    """
    Stop a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        None: If successful.
        dict[str, str]: An error dictionary if an error occurs.
    """
    try:
        await stop_lab(lid)
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
async def wipe_cml_lab(lid: UUID4Type, ctx: Context) -> None | dict[str, str]:
    """
    Wipe a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        None: If successful.
        dict[str, str]: An error dictionary if an error occurs.
    """
    try:
        result = await ctx.elicit("Are you sure you want to wipe the lab?")
        if result.action == "accept":
            await wipe_lab(lid)
        else:
            raise Exception("Wipe operation cancelled by user.")
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
async def delete_cml_lab(lid: UUID4Type, ctx: Context) -> None | dict[str, str]:
    """
    Delete a CML lab by its ID. If the lab is running and/or not wiped, it will be stopped and wiped first.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        None: If successful.
        dict[str, str]: An error dictionary if an error occurs.
    """
    try:
        result = await ctx.elicit("Are you sure you want to delete the lab?")
        if result.action == "accept":
            await stop_lab(lid)  # Ensure the lab is stopped before deletion
            await wipe_lab(lid)  # Ensure the lab is wiped before deletion
            await cml_client.delete(f"/labs/{lid}")
        else:
            raise Exception("Delete operation cancelled by user.")
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
async def configure_cml_node(lid: UUID4Type, nid: UUID4Type, config: NodeConfigurationContent) -> UUID4Type | dict[str, str]:
    """
    Configure a node in a CML lab by its lab ID and node ID. The node must be in a BOOTED (i.e., wiped) state.

    Args:
        lid (UUID4Type): The lab ID.
        nid (UUID4Type): The node ID.
        config (NodeConfigurationContent): The configuration content for the node.

    Returns:
        UUID4Type: The ID of the node configured if successful.
        dict[str, str]: An error dictionary if an error occurs.
    """
    payload = {"configuration": str(config)}
    try:
        resp = await cml_client.patch(f"/labs/{lid}/nodes/{nid}", data=payload)
        return UUID4Type(resp)
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
async def get_nodes_for_cml_lab(lid: UUID4Type) -> list[Node] | dict[str, str]:
    """
    Get a list of nodes for a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        list[Node]: List of node objects (without their configurations).
        dict[str, str]: An error dictionary if an error occurs.
    """
    try:
        resp = await cml_client.get(f"/labs/{lid}/nodes", data={"data": True, "operational": True, "exclude_configurations": True})
        return [Node(**node) for node in resp]
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
async def get_cml_lab_by_title(title: LabTitle) -> Lab | dict[str, str]:
    """
    Get a CML lab by its title.

    Args:
        title (LabTitle): The lab title.

    Returns:
        Lab: The lab object if found.
        dict[str, str]: An error dictionary if an error occurs or lab is not found.
    """
    try:
        labs = await get_all_labs()
        for lid in labs:
            lab = await cml_client.get(f"/labs/{lid}")
            if lab["lab_title"] == str(title):
                return Lab(**lab)
        return {"error": "Lab not found"}
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
async def stop_cml_node(lid: UUID4Type, nid: UUID4Type) -> None | dict[str, str]:
    """
    Stop a node in a CML lab by its lab ID and node ID.

    Args:
        lid (UUID4Type): The lab ID.
        nid (UUID4Type): The node ID.

    Returns:
        None: If successful.
        dict[str, str]: An error dictionary if an error occurs.
    """
    try:
        await cml_client.put(f"/labs/{lid}/nodes/{nid}/state/stop")
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
async def start_cml_node(lid: UUID4Type, nid: UUID4Type) -> None | dict[str, str]:
    """
    Start a node in a CML lab by its lab ID and node ID.

    Args:
        lid (UUID4Type): The lab ID.
        nid (UUID4Type): The node ID.

    Returns:
        None: If successful.
        dict[str, str]: An error dictionary if an error occurs.
    """
    try:
        await cml_client.put(f"/labs/{lid}/nodes/{nid}/state/start")
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
async def wipe_cml_node(lid: UUID4Type, nid: UUID4Type, ctx: Context) -> None | dict[str, str]:
    """
    Wipe a node in a CML lab by its lab ID and node ID. Node must be stopped first.

    Args:
        lid (UUID4Type): The lab ID.
        nid (UUID4Type): The node ID.

    Returns:
        None: If successful.
        dict[str, str]: An error dictionary if an error occurs.
    """
    try:
        result = await ctx.elicit("Are you sure you want to wipe the node?")
        if result.action == "accept":
            await cml_client.put(f"/labs/{lid}/nodes/{nid}/wipe_disks")
        else:
            raise Exception("Wipe operation cancelled by user.")
    except httpx.HTTPStatusError as e:
        return {"error": f"HTTP error {e.response.status_code}: {e.response.text}"}
    except Exception as e:
        return {"error": str(e)}


@server_mcp.tool
def send_cli_command(lid: UUID4Type, label: NodeLabel, command: str, config_command: bool = False) -> str | dict[str, str]:
    """
    Send a CLI command to a node in a CML lab by its lab ID and node label.

    Args:
        lid (UUID4Type): The lab ID.
        label (NodeLabel): The label of the node to send the command to.
        command (str): The CLI command to send.
        config_command (bool, optional): If True, send as a configuration command. Defaults to False.

    Returns:
        str | dict[str, str]: The command output if successful, or a dictionary with an error message if an error occurs.
    """
    cwd = os.getcwd()  # Save the current working directory
    try:
        os.chdir(tempfile.gettempdir())  # Change to a writable directory (required by pyATS/ClPyats)
        lab = cml_client.vclient.join_existing_lab(str(lid))  # Join the existing lab using the provided lab ID
        pylab = ClPyats(lab)  # Create a ClPyats object for interacting with the lab
        pylab.sync_testbed(settings.cml_username, settings.cml_password)  # Sync the testbed with CML credentials
        if config_command:
            # Send the command as a configuration command
            return pylab.run_config_command(str(label), command)
        # Send the command as an exec/operational command
        return pylab.run_command(str(label), command)
    except Exception as e:
        return {"error": str(e)}
    finally:
        os.chdir(cwd)  # Restore the original working directory



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from typing import Annotated

from pydantic import BaseModel, Field, conlist

from cml_mcp.schemas.common import Hostname, IPAddress, UUID4Type

Hostpubkey = Annotated[
    str,
    Field(pattern=r"^[a-zA-Z\d_-]{1,40}(?:@openssh.com)? [a-zA-Z0-9+/=]{1,16000}$"),
]


# $id: http://virl.lab/compute_host.json
class ComputeHostInternal(BaseModel, extra="forbid"):
    id: UUID4Type = Field(...)
    server_address: IPAddress = Field(...)
    hostname: Hostname = Field(...)
    is_simulator: bool = Field(...)
    is_connector: bool = Field(...)
    sshd_host_keys: conlist(Hostpubkey, min_length=0, max_length=5) = None



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
import re

from pydantic import BaseModel, ConfigDict, Field

from cml_mcp.schemas.common import DefinitionID, FilePath
from cml_mcp.schemas.nodes import NodeConfigurationContent


class Pyats(BaseModel, extra="forbid"):
    username: str | None = Field(
        default=None,
        min_length=1,
        max_length=64,
        description="The pyATS username to be used with this image.",
    )
    password: str | None = Field(
        default=None,
        min_length=1,
        max_length=128,
        description="The pyATS password to be used with this image.",
    )


class ImageDefinition(BaseModel, extra="forbid"):
    id: DefinitionID = Field(..., description="The identifier of this image definition.")
    node_definition_id: DefinitionID = Field(..., description="Node definition ID for the image definition.")
    label: str = Field(
        ...,
        description="A required label for the image definition.",
        min_length=1,
        max_length=64,
    )
    disk_image: FilePath = Field(..., description="A source image for the image definition.")
    efi_boot: bool = Field(
        default=None,
        description="Whether to use EFI for booting.",
    )
    sha256: str = Field(
        default=None,
        description="SHA256 of the disk_image (optional)",
        examples=["58ce6f1271ae1c8a2006ff7d3e54e9874d839f573d8009c20154ad0f2fb0a225"],
        min_length=64,
        max_length=64,
        pattern=re.compile(r"^[a-fA-F\d]{64}(?![\n\r])$"),
    )
    schema_version: str = Field(
        default=None,
        description="The image definition schema version.",
        examples=["0.0.1"],
        min_length=1,
        max_length=32,
    )
    description: str = Field(
        default=None,
        min_length=0,
        max_length=4096,
        description="An optional description for the image definition.",
    )
    disk_image_2: FilePath = Field(default=None, description="A second source image for the image definition.")
    disk_image_3: FilePath = Field(default=None, description="A third source image for the image definition.")
    disk_image_4: FilePath = Field(default=None, description="A fourth source image for the image definition.")
    read_only: bool = Field(
        default=None,
        description="Whether the image definition can be updated or deleted.",
    )
    # the 20 MB limit in nginx is not set for node definitions
    # (the default 100 kB limit is set instead)
    configuration: NodeConfigurationContent = None
    pyats: Pyats | None = None
    ram: int | None = Field(default=None, ge=1, le=1048576, description="Memory (MiB).")
    cpus: int | None = Field(default=None, ge=1, le=128, description="CPUs.")
    cpu_limit: int | None = Field(default=None, ge=20, le=100, description="CPU Limit.")
    data_volume: int | None = Field(default=None, ge=0, le=4096, description="Data Disk Size (GiB).")
    boot_disk_size: int | None = Field(default=None, ge=0, le=4096, description="Boot Disk Size (GiB).")
    # TODO: verify it works as expected not tested yet
    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "id": "alpine-3-10-base",
                "node_definition_id": "lxc",
                "description": "Alpine Linux and network tools",
                "label": "Alpine 3.10",
                "disk_image": "alpine-3-10-base.qcow2",
                "read_only": True,
                "ram": None,
                "cpus": None,
                "data_volume": None,
                "boot_disk_size": None,
            }
        }
    )



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
import re
from typing import Annotated, Literal

from fastapi import Body
from pydantic import BaseModel, Field, model_validator

from cml_mcp.schemas.common import (
    DOMAIN_REG,
    PORT_REG,
    GroupName,
    UserName,
    UUID4Type,
)

from .users import Password


class UserAuthData(BaseModel):
    """Name and password of the user to be tested."""

    username: UserName = Field(...)
    password: Password = Field(...)


AuthDataBody = Annotated[
    UserAuthData,
    Body(
        description="This request body is a JSON object that holds authentication data."
    ),
]


CNArray = Annotated[
    list[str],
    Field(description="An array of LDAP CNs.", examples=["group-1", "group-2"]),
]

# length validation moved to backend so it does not return entered certificate
# in response on unsuccessful validation
PublicCertificate = Annotated[
    str,
    Field(
        description="Contents of the public certificate in PEM format.",
        examples=[
            """
                -----BEGIN CERTIFICATE-----
                MIIDGzCCAgOgAwIBAgIBATANBgkq
                gno5gnopebgtAOFFHUnrr35n52/4
                //shortened
                xlJQaTOM9rpsuO/Q==
                -----END CERTIFICATE-----
                """
        ],
    ),
]

JwtToken = Annotated[
    str,
    Field(
        description="JWT token",
        pattern="^[A-Za-z0-9-_]+.[A-Za-z0-9-_]+.[A-Za-z0-9-_]+$",
        examples=[
            (
                "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJjb20uY2lzY28udmlybCIsI"
                "mlhdCI6MTYxMzc1MDgxNSwiZXhwIjoxNjEzODM3MjE1LCJzdWIiOiIwNDg5MDcxNS00YWE"
                "3LTRhNDAtYWQzZS1jZThmY2JkNGQ3YWEifQ.Q4heV5TTYQ6yhpJ5GKLm_Bf9D9NL-wDxL9"
                "Orz1ByxWs"
            )
        ],
    ),
]


class GroupAuthData(BaseModel):
    """Name of the group to be tested."""

    group_name: GroupName = Field(...)


class SystemAuthConfigBase(BaseModel, extra="allow"):
    """System Authentication Configuration Object."""

    method: Literal["ldap", "local"] = Field(
        ..., description="What authentication method should be used."
    )
    # allow empty string or string matches pattern
    server_urls: str = Field(
        default=None,
        max_length=256,
        description="URI of LDAP server, either LDAP or LDAPS,"
        "multiple servers can be specified, separate with space.",
        examples=["ldaps://ad.corp.com:3269"],
        pattern=re.compile(
            rf"^(?:(?: (?=l))?ldaps?://(?:{DOMAIN_REG})(?::{PORT_REG})?)*(?![\n\r])$"
        ),
    )
    verify_tls: bool = Field(
        default=None,
        description="Set to `false` if certificates should not be verified.",
    )
    cert_data_pem: PublicCertificate = Field(
        default=None, description="Reference to a public certificate."
    )
    use_ntlm: bool = Field(
        default=None,
        description="If `true` then password for manager user"
        " will be stored as NTLM hash. "
        "Only works with ActiveDirectory servers.",
    )
    root_dn: str = Field(
        default=None,
        max_length=256,
        description="The root DN that will be applied.",
        examples=["DC=corp,DC=com"],
    )
    user_search_base: str = Field(
        default=None,
        max_length=256,
        description="The user search base where users should be looked up. "
        "Typically a OU or CN. Will be combined with the root DN.",
        examples=["CN=users,CN=accounts"],
    )
    user_search_filter: str = Field(
        default=None,
        max_length=1024,
        description="The filter that will be applied to the user. "
        "Must have a placeholder `{0}` replaced with the username.",
        examples=[
            "(&(uid={0})(memberOf=CN=cmlusers,CN=groups,CN=accounts,DC=corp,DC=com))"
        ],
    )
    admin_search_filter: str = Field(
        default=None,
        max_length=1024,
        description="Same as for the user search filter. "
        "Grants admin rights if matched.",
        examples=[
            "(&(uid={0})(memberOf=CN=cmladmins,CN=groups,CN=accounts,DC=corp,DC=com))"
        ],
    )
    group_search_base: str = Field(
        default=None,
        max_length=256,
        description="The group search base where groups should be looked up. "
        "Typically a OU or CN. Will be combined with the root DN.",
        examples=["CN=groups,CN=accounts"],
    )
    group_search_filter: str = Field(
        default=None,
        max_length=1024,
        description="The filter applied to groups. "
        "Must have a placeholder `{0}` replaced with the group name.",
        examples=["(&(cn={0})(objectclass=posixgroup))"],
    )
    group_via_user: bool = Field(
        default=None,
        description="If `true`, use `group_user_attribute` "
        "to determine user group memberships.",
    )
    group_user_attribute: str = Field(
        default=None,
        max_length=64,
        description="Attribute of the user that holds group memberships.",
        examples=["memberOf"],
    )
    group_membership_filter: str = Field(
        default=None,
        max_length=1024,
        description="Filter to apply to groups specifying the user.",
        examples=["(member={0})"],
    )
    manager_dn: str = Field(
        default=None,
        max_length=256,
        description="Manager user DN for lookup if anonymous search is not allowed.",
        examples=["uid=someuser,cn=users,cn=accounts,dc=corp,dc=com"],
    )
    display_attribute: str = Field(
        default=None,
        max_length=256,
        description="User attribute for displaying the logged in user.",
        examples=["displayName"],
    )
    group_display_attribute: str = Field(
        default=None,
        max_length=256,
        description="Group attribute for displaying group description.",
        examples=["description"],
    )
    email_address_attribute: str = Field(
        default=None,
        max_length=64,
        description="User attribute for displaying the email address.",
        examples=["mail"],
    )
    resource_pool: UUID4Type | None = Field(
        default=None, description="Resource pool or template ID for new user accounts."
    )


class SystemAuthConfigRequest(SystemAuthConfigBase, extra="forbid"):
    manager_password: str = Field(
        default=None,
        max_length=256,
        description="""
          The password for the management user. If `use_ntlm` is `true` then the
          password will be converted to a NTLM hash and the hash is stored.
          Otherwise, the cleartext password will be stored using obfuscation.
          """,
    )


class SystemAuthConfigResponse(SystemAuthConfigBase, extra="forbid"):
    pass


class SystemAuthTestData(BaseModel, extra="forbid"):
    """
    Test Configuration Data which includes the system authentication
    configuration, and either a username and a password, or a group name.
    """

    auth_config: SystemAuthConfigRequest = Field(..., alias="auth-config")
    auth_data: UserAuthData = Field(default=None, alias="auth-data")
    group_data: GroupAuthData = Field(default=None, alias="group-data")

    @model_validator(mode="after")
    def check_any_of(self):
        if self.auth_data is None and self.group_data is None:
            raise ValueError("Either auth_data or group_data must be provided.")
        return self


class AuthTestAttributes(BaseModel, extra="allow"):
    pass


class AuthTestUserResponse(BaseModel):
    auth_ok: bool = Field(
        default=None,
        description="The user has access to the system and the user filter matches.",
    )
    is_admin: bool = Field(
        default=None,
        description="The user has admin rights. If LDAP is configured, "
        "then the admin filter must match.",
    )
    display: str = Field(
        default=None,
        description="The user's display name, if the configured attribute was found.",
    )
    email: str = Field(
        default=None,
        description="The user's email address, if the configured attribute was found.",
    )
    attributes: AuthTestAttributes = Field(
        default=None,
        description=(
            "If the user auth'ed OK, then this object holds the dictionary of user attributes, "
            "independent of having CML access or admin access or other privileges. "
            "If an error occurred (config/server problem), then the error message will be included "
            "in the attributes."
        ),
    )


class AuthTestGroupResponse(BaseModel):
    group_ok: bool = Field(
        default=None,
        description="The group exists on the server and the group filter matches.",
    )
    is_member: bool = Field(
        default=None,
        description=(
            "Whether the given user is a member of the given group. Will default to false "
            "if either the user or the group could not be found on the LDAP server."
        ),
    )
    display: str = Field(
        default=None,
        description="The group's display name, if the configured attribute was found.",
    )
    attributes: AuthTestAttributes = Field(
        default=None,
        description=(
            "If the group was found OK, then this object holds the dictionary of group attributes, "
            "independent of having CML access or admin access or other privileges. "
            "If an error occurred (config/server problem), then the error message will be included "
            "in the attributes."
        ),
    )


class AuthTestResponse(BaseModel, extra="forbid"):
    """System Authentication Test response."""

    user: AuthTestUserResponse = Field(
        default=None, description="Results for the user."
    )
    group: AuthTestGroupResponse = Field(
        default=None, description="Results for the group."
    )


class AuthenticateResponse(BaseModel):
    """Authenticate response."""

    username: str = Field(default=None, examples=["admin"])
    id: UUID4Type = Field(default=None, description="ID of a user")
    token: JwtToken = Field(default=None)
    admin: bool = Field(default=None, examples=[False])
    # connection returns null when no errors
    error: str | None = Field(
        default=None,
        description="Error messages for errors that occurred while authenticating, "
        "but did not interrupt the login, such as LDAP group membership "
        "refresh errors.",
        examples=[
            "Could not refresh LDAP group memberships "
            "(Invalid base DN or root DN format)"
        ],
    )



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from typing import Annotated, Literal

from fastapi import Body
from pydantic import BaseModel, Field

from cml_mcp.schemas.common import (
    Label,
    LinuxInterfaceName,
    MACAddress,
    State,
    UUID4Type,
)

InterfaceLabel = Annotated[Label, Field(..., description="An interface label.")]
InterfaceSlot = Annotated[int, Field(ge=0, le=128, description="Number of slots.")]

InterfaceType = Annotated[Literal["physical", "loopback"], Field(description="Interface type.")]


class InterfaceCreate(BaseModel, extra="forbid"):
    node: UUID4Type = Field(...)
    slot: InterfaceSlot = Field(default=None)
    mac_address: MACAddress = Field(default=None)


class InterfaceUpdate(BaseModel, extra="forbid"):
    mac_address: MACAddress = Field(default=None)


# without id
class InterfaceBase(BaseModel):
    """Interface object."""

    lab_id: UUID4Type = Field(default=None, description="ID of the lab.")
    label: InterfaceLabel = Field(default=None)
    type: InterfaceType = Field(default=None)
    slot: InterfaceSlot = Field(default=None)
    mac_address: MACAddress = Field(default=None)
    is_connected: bool = Field(default=None, description="Whether this interface is connected (in-use).")
    state: State = Field(default=None, description="The status of the link in the lab.")


InterfaceCreateBody = Annotated[
    InterfaceCreate,
    Body(
        description="A JSON object that specifies a request to create an interface "
        "on a node. If the slot is omitted, the request indicates "
        "the *first* unused slot on the node. If the slot is specified, "
        "the request indicates *all unallocated* slots up to and including "
        "that slot."
    ),
]

InterfaceUpdateBody = Annotated[
    InterfaceUpdate,
    Body(description="A JSON object with an interface's updatable properties."),
]


# responses
class InterfaceOperationalDataResponse(BaseModel, extra="forbid"):
    device_name: LinuxInterfaceName | None = Field(default=None, max_length=64, description="Device name.")
    mac_address: MACAddress = Field(default=None)
    src_udp_port: int | None = Field(default=None, ge=0, le=65535, description="Source UDP port.")
    dst_udp_port: int | None = Field(default=None, ge=0, le=65535, description="Destination UDP port.")


class InterfaceResponse(InterfaceBase, extra="forbid"):
    """The response body is a JSON interface object."""

    id: UUID4Type = Field(..., description="ID of the interface.")
    label: InterfaceLabel = Field(...)
    node: UUID4Type = Field(default=None, description="ID of the node.")
    src_udp_port: int | None = Field(default=None, ge=0, le=65535, description="Source UDP port (operational).")
    dst_udp_port: int | None = Field(default=None, ge=0, le=65535, description="Destination UDP port (operational).")
    device_name: LinuxInterfaceName | None = Field(default=None, max_length=64, description="Device name (operational).")
    operational: InterfaceOperationalDataResponse = Field(
        default=None,
        description="Additional operational data associated with the interface.",
    )
    slot: InterfaceSlot | None = Field(default=None)


class InterfaceStateResponse(BaseModel, extra="forbid"):
    label: InterfaceLabel = Field(...)
    state: State = Field(...)



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from enum import StrEnum
from typing import Annotated

from fastapi import Body
from pydantic import BaseModel, ConfigDict, Field, model_validator

from cml_mcp.schemas.common import (
    Coordinate,
    DefinitionID,
    Label,
    PinnedComputeID,
    TagArray,
    UUID4Type,
)

NodeLabel = Annotated[Label, Field(..., description="A node label.", examples=["desktop-1"])]
NodeId = Annotated[
    UUID4Type,
    Field(
        ...,
        description="A node UUID4.",
        examples=["26f677f3-fcb2-47ef-9171-dc112d80b54f"],
    ),
]

Ram = Annotated[
    int | None,
    Field(
        ge=1,
        le=1048576,
        description="RAM size in MB. Can be null.",
    ),
]

Cpus = Annotated[int | None, Field(ge=1, le=128, description="Number of CPUs. Can be null.")]

AllocatedCpus = Annotated[
    int | None,
    Field(ge=0, le=128, description="Number allocated of CPUs. Can be null."),
]

CpuLimit = Annotated[int | None, Field(ge=20, le=100, description="CPU limit percentage. Can be null.")]

DiskSpace = Annotated[int | None, Field(ge=0, le=4096, description="Disk space in GB. Can be null.")]

IOLAppId = Annotated[int | None, Field(ge=1, le=1022, description="IOL Application ID. Can be null.")]

# max length is 20MB and is checked by nginx
# it makes no sense to validate it here as it mostly just pollutes logs
NodeConfigurationContent = Annotated[str | None, Field(description="Node configuration (no more than 20MB).")]


class NodeConfigurationFile(BaseModel, extra="forbid"):
    name: Annotated[
        str,
        Field(
            ...,
            min_length=1,
            max_length=64,
            description=(
                "The name of the configuration file. Can also use the keyword "
                '"Main" to denote the main configuration file for the given definition.'
            ),
        ),
    ]
    content: NodeConfigurationContent = Field(default=None)


# When GET, we need to both be backwards compatible and return all files, which
# is only possible by deprecating the old `/configuration` parameter and adding a new
# `/configurations` parameter which returns an array of files.
NodeConfigurationFiles = Annotated[
    list[NodeConfigurationFile],
    Field(description="List of node configuration file objects."),
]

# When PATCHing, all these formats make sense; just a string if all you want to touch
# is the main config, a single object if you just want to edit one non-main config,
# or an array if you want to do multiple files at once.
NodeConfiguration = Annotated[
    NodeConfigurationContent | NodeConfigurationFiles | NodeConfigurationFile,
    Field(
        description=(
            "Node configuration. Either an array of file objects, or a single file "
            "object, or just the content of the main configuration file."
        )
    ),
]


class NodeParameters(BaseModel, extra="allow"):
    model_config = ConfigDict(
        json_schema_extra={
            "description": "Key-value pairs of a custom node SMBIOS parameters.",
            "example": {"smbios.bios.vendor": "Lenovo"},
        }
    )

    @model_validator(mode="after")
    def check_types(self):
        for field_name, value in self.__dict__.items():
            # basically just nullable string.
            if value is not None and not isinstance(value, str):
                raise TypeError(f"Value for '{field_name}' must be a string or None," f"but got {type(value).__name__}.")
        return self


# Shared properties for all Node objects, excluding configuration
class NodeBase(BaseModel, extra="allow"):
    """Node object base."""

    x: Coordinate = Field(default=None, description="Node X coordinate.")
    y: Coordinate = Field(default=None, description="Node Y coordinate.")
    label: NodeLabel = Field(default=None)
    parameters: NodeParameters = Field(default=None)
    image_definition: DefinitionID | None = Field(default=None, description="Image Definition ID for the specified node.")
    ram: Ram = Field(default=None)
    cpu_limit: CpuLimit = Field(default=None)
    data_volume: DiskSpace = Field(default=None)
    boot_disk_size: DiskSpace = Field(default=None)
    hide_links: bool | None = Field(default=None, description="Whether to hide links to/from this node.")
    tags: TagArray = Field(default=None)


# Exactly the base properties are updatable, but we can also use string to set just
# the main config file or a single object outside an array
class NodeUpdate(NodeBase, extra="forbid"):
    cpus: Cpus = Field(default=None)
    configuration: NodeConfiguration = Field(default=None)
    pinned_compute_id: PinnedComputeID = Field(default=None)


class NodeStates(StrEnum):
    DEFINED_ON_CORE = "DEFINED_ON_CORE"
    STOPPED = "STOPPED"
    STARTED = "STARTED"
    QUEUED = "QUEUED"
    BOOTED = "BOOTED"
    DISCONNECTED = "DISCONNECTED"


class BootProgresses(StrEnum):
    NOT_RUNNING = "Not running"
    BOOTING = "Booting"
    BOOTED = "Booted"


BootProgress = Annotated[BootProgresses, Field(description="Node boot progress.")]


# Node requires node definition, position and label
class NodeDefinedBase(NodeBase, extra="allow"):
    node_definition: DefinitionID = Field(..., description="Node Definition ID for the specified node.")
    x: Coordinate = Field(...)
    y: Coordinate = Field(...)
    label: NodeLabel = Field(...)


class NodeDefined(NodeDefinedBase, extra="allow"):
    cpus: Cpus = Field(default=None)


# Create node requires node definition, position and label
class NodeCreate(NodeDefined, extra="forbid"):
    configuration: NodeConfiguration = Field(default=None)


class ConsoleKeyDetails(BaseModel, extra="forbid"):
    console_key: UUID4Type = Field(default=None)
    label: Label = Field(default=None)
    device_number: int = Field(default=None)


class NodeOperationalData(BaseModel, extra="forbid"):
    """Operational data resolved when a node was started."""

    boot_disk_size: DiskSpace = Field(...)
    cpu_limit: CpuLimit = Field(...)
    cpus: AllocatedCpus = Field(...)
    data_volume: DiskSpace = Field(...)
    ram: Ram = Field(...)
    compute_id: UUID4Type | None = Field(..., description="The ID of the compute host where this node is deployed.")
    image_definition: DefinitionID | None = Field(..., description="Image definition ID used for the specified node.")
    vnc_key: UUID4Type | None = Field(
        ...,
        description="The key used to connect to a node's graphical VNC console " "if supported by node.",
    )
    resource_pool: UUID4Type | None = Field(
        default=None,
        description="Node was launched with resources from the given resource pool.",
    )
    iol_app_id: IOLAppId = Field(default=None)
    serial_consoles: list[ConsoleKeyDetails] = Field(default=None)


# Node with operational data and proper UUID id
class Node(NodeDefinedBase, extra="forbid"):
    id: NodeId
    boot_progress: BootProgress = Field(
        default=None,
        description="Flag indicating whether the node appears to have completed its boot.",
    )
    compute_id: UUID4Type | None = Field(
        default=None,
        description="The ID of the compute host where this node is deployed.",
    )
    cpus: AllocatedCpus = Field(...)
    iol_app_id: IOLAppId = Field(default=None)
    operational: NodeOperationalData = Field(
        default=None,
        description="Additional operational data associated with the node.",
    )
    resource_pool: UUID4Type | None = Field(
        default=None,
        description="Node was launched with resources from the given resource pool.",
    )
    state: NodeStates = Field(default=None, description="The state of the node.")
    vnc_key: UUID4Type | None = Field(
        default=None,
        description="The key used to connect to a node's graphical VNC console, if supported by node.",
    )
    configuration: NodeConfiguration = Field(default=None)
    pinned_compute_id: PinnedComputeID = Field(default=None)
    lab_id: UUID4Type = Field(default=None)
    serial_consoles: list[ConsoleKeyDetails] = Field(default=None)


NodeCreateBody = Annotated[NodeCreate, Body(description="A JSON object with a node's fundamental properties.")]

NodeUpdateBody = Annotated[NodeUpdate, Body(description="A JSON object with a node's updatable properties.")]


# responses
class NodeStateResponse(BaseModel, extra="forbid"):
    state: str = Field(default=None)
    progress: str = Field(default=None)


NodeResponse = Annotated[Node, Field(description="The response body is a JSON node object.")]



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
import re
from datetime import datetime
from typing import Annotated

from fastapi import Body
from pydantic import BaseModel, Field, conlist

from cml_mcp.schemas.common import DOMAIN_REG, HTTP_URL_REG, Hostname


class LicensingTimeInfo(BaseModel):
    """Timing information on communication with SSMS."""

    succeeded: datetime | None = Field(
        ...,
        description="The time when the given request last completed with success.",
    )
    attempted: datetime | None = Field(..., description="The time when the given request was last made.")
    scheduled: datetime | None = Field(
        ...,
        description="The time when the given request will be made next without intervention.",
    )
    status: str | None = Field(..., description="The status result of the last attempt.")
    failure: str | None = Field(..., description="The failure reason of the last attempt.")
    success: str | None = Field(..., description="The status of the last communication attempt.")


class LicensingFeatureIdMixIn(BaseModel):
    """Individually licensed feature description."""

    id: str = Field(
        ...,
        description="Identification tag of the feature.",
        pattern=re.compile(r"^[a-zA-Z\d._,-]{1,128}(?![\n\r])$"),
        examples=["regid.2019-10.com.cisco.CML_NODE_COUNT,1.0_2607650b-6ca8-46d5-81e5-e6688b7383c4"],
    )


LicensingProductLicenseBodyParameter = Annotated[
    str,
    Body(
        description="Product license.",
        pattern=re.compile(r"^\w{1,20}(?![\n\r])$"),
    ),
]


class LicensingTransportProxy(BaseModel, extra="forbid"):
    """HTTP Proxy to contact SSMS."""

    server: str | None = Field(
        ...,
        description="Domain name of the HTTP proxy server.",
        examples=["lab-proxy.example.com"],
        pattern=re.compile(rf"^({DOMAIN_REG})(?![\n\r])$"),
        min_length=1,
        max_length=256,
    )
    port: int | None = Field(
        ...,
        description="Port of the HTTP proxy server.",
        ge=1,
        le=65535,
        examples=[80],
    )


class Udi(BaseModel, extra="forbid"):
    hostname: Hostname = Field(
        default=None,
        description="Hostname of this product instance within SSMS associated with registration.",
    )
    product_uuid: str = Field(
        default=None,
        description="ID of this product instance within SSMS associated with registration.",
    )


class Registration(BaseModel, extra="forbid"):
    status: str = Field(
        default=None,
        description="The current registration status of this product instance.",
    )
    smart_account: str | None = Field(
        default=None,
        description="Name of the customer Smart Account associated with registration.",
    )
    virtual_account: str | None = Field(
        default=None,
        description="Name of the virtual sub-account associated with registration.",
    )
    register_time: LicensingTimeInfo = Field(default=None)
    renew_time: LicensingTimeInfo = Field(default=None)
    expires: datetime | None = Field(
        default=None,
        description="The time current valid registration is due to expire.",
    )


class Authorization(BaseModel, extra="forbid"):
    status: str = Field(
        default=None,
        description="The current authorization status of this product instance.",
    )
    renew_time: LicensingTimeInfo = Field(default=None)
    expires: datetime | None = Field(
        default=None,
        description="The time current valid authorization is due to expire.",
    )


class ProductLicense(BaseModel, extra="forbid"):
    active: str = Field(default=None, description="Currently active product license.")
    is_enterprise: bool = Field(
        default=None,
        description="Whether the active product license includes enterprise features.",
    )


class LicensingFeature(LicensingFeatureIdMixIn, extra="forbid"):
    name: str = Field(default=None, description="Short name of the feature.")
    description: str = Field(default=None, description="Description of the feature.")
    version: str = Field(default=None, description="Version of the feature.")
    in_use: int = Field(
        default=None,
        description="Currently requested count of uses for this feature.",
    )
    status: str = Field(
        default=None,
        description="Current authorization status for this individual feature.",
    )
    min: int = Field(default=None, description="The minimal count for this individual feature.", ge=0)
    max: int = Field(default=None, description="The maximal count for this individual feature.", ge=0)
    minEndDate: datetime | None = Field(
        default=None,
        description="First date in which a valid license reservation expires.",
    )
    maxEndDate: datetime | None = Field(
        default=None,
        description="Last date in which a valid license reservation expires.",
    )


class LicensingTransport(BaseModel, extra="forbid"):
    """Configuration for Smart Licensing transport."""

    proxy: LicensingTransportProxy = Field(...)
    ssms: str | None = Field(
        default="https://smartreceiver.cisco.com/licservice/license",
        description="The URL.",
        pattern=re.compile(HTTP_URL_REG),
        min_length=1,
        max_length=256,
        examples=["https://ssms-satellite.example.com:8443/Transportgateway/services/DeviceRequestHandler"],
    )


class Transport(LicensingTransport):
    default_ssms: str = Field(
        ...,
        description="The main production URL which shall be set unless changed by user.",
    )


class LicensingStatus(BaseModel, extra="forbid"):
    """Configuration and status of Smart Licensing."""

    udi: Udi = Field(default=None, description="The product instance identifier.")
    registration: Registration = Field(default=None, description="Product registration status.")
    authorization: Authorization = Field(default=None, description="Product overall feature authorization status.")
    reservation_mode: bool = Field(default=None, description="The current reservation mode status.")
    features: list[LicensingFeature] = Field(default=None)
    product_license: ProductLicense = Field(default=None)
    transport: Transport = Field(default=None)


class LicensingFeatureCount(LicensingFeatureIdMixIn, extra="forbid"):
    count: int = Field(..., description="Requested count of this feature.", ge=0)


LicensingFeatureCounts = Annotated[
    conlist(LicensingFeatureCount, min_length=1, max_length=2),
    Field(description="List of individual feature explicit counts."),
]


class LicensingRegistration(BaseModel, extra="forbid"):
    token: str = Field(
        ...,
        description="A token generated by the target SSMS instance to authorize product to it.",
        pattern=r"^[a-zA-Z\d+/%=-]{1,256}$",
    )
    reregister: bool = Field(default=False, description="Request reregistration from the current SSMS.")


LicensingRegistrationBody = Annotated[LicensingRegistration, Body(...)]



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
import re
from datetime import UTC, datetime
from enum import StrEnum
from typing import Annotated

from pydantic import AfterValidator, BaseModel, Field, field_serializer

from cml_mcp.schemas.simple_common.models import DefaultPermissions

LINT_REG = r"[\da-z-]{1,15}"
UUID4_REG = (
    r"^[\da-f]{8}-[\da-f]{4}-4[\da-f]{3}-[89ab][\da-f]{3}-[\da-f]{12}(?![\n\r])$"
)

IPV4_REG = r"(\d{1,3}.){3}\d{1,3}"
IPV6_REG = r"[\da-fA-F:]{3,39}" + f"(%{LINT_REG})?"
IP_REG = rf"^({IPV4_REG}|{IPV6_REG})(?![\n\r])$"
HOSTNAME_REG = r"[a-zA-Z\d.-]{1,64}"
PORT_REG = r"\d{1,5}"
HOST_REG = rf"^{HOSTNAME_REG}(?![\n\r])$"
DOMAIN_REG = rf"{IPV4_REG}|\[{IPV6_REG}\]|{HOSTNAME_REG}"
ORIGIN_REG = rf"https?://({DOMAIN_REG})(:{PORT_REG})?"
HTTP_URL_REG = rf"^{ORIGIN_REG}(/[\w.-]+)+(?![\n\r])$"


def _remove_duplicates(lst: list) -> list:
    # list elements must be hashable
    return list(set(lst))


FilePath = Annotated[
    str,
    Field(
        min_length=1,
        max_length=255,
        pattern=re.compile(r"^(?![.])[^!@#%^&*();$\n\r\t/\\]{1,255}(?![\n\r])$"),
    ),
]

# Create a custom annotated string for UUID4 as using UUID4 from Pydantic
# causes issues with SQLAlchemy. We need to wrap fields with str(lab_id)
# because passing the UUID4 class directly does not work.
# A UUID-4 has five groups of lowercase hexadecimal characters, the first has 8 chars,
# the second 4 chars, the third 4 chars, the fourth 4 chars, the fifth 12 chars.
# However, to make it a valid UUID4 the third group (the one in the middle)
# must start with a 4 and the fourth group must start with 8, 9, a or b.
UUID4Type = Annotated[
    str,
    Field(
        description="A UUID4",
        examples=["90f84e38-a71c-4d57-8d90-00fa8a197385"],
        pattern=re.compile(UUID4_REG),
    ),
]


UUID4ArrayType = Annotated[
    list[UUID4Type],
    Field(
        examples=[
            [
                "90f84e38-a71c-4d57-8d90-00fa8a197385",
                "60f84e39-ffff-4d99-8a78-00fa8aaf5666",
            ]
        ],
    ),
]


GenericDescription = Annotated[str, Field(max_length=4096)]

NullableGenericDescription = Annotated[str | None, Field(max_length=4096)]

UserName = Annotated[
    str,
    Field(
        description="The name of the user.",
        examples=["admin"],
        min_length=1,
        max_length=32,
    ),
]


UserFullName = Annotated[
    str,
    Field(
        description="The full name of the user.",
        examples=["Dr. Super User"],
        max_length=128,
    ),
]

GroupName = Annotated[
    str,
    Field(
        description="The full name of the group.",
        examples=["CCNA Study Group Class of 21"],
        min_length=1,
        max_length=64,
    ),
]

IPAddress = Annotated[
    str,
    Field(
        pattern=re.compile(IP_REG),
        description="An IPv4 or IPv6 host address.",
    ),
]

IPv4Address = Annotated[
    str,
    Field(
        pattern=re.compile(IPV4_REG),
        description="An IPv4 host address.",
    ),
]

IPv6Address = Annotated[
    str,
    Field(
        pattern=re.compile(IPV6_REG),
        description="An IPv6 host address.",
    ),
]

IPNetwork = Annotated[
    str,
    Field(
        pattern=re.compile(rf"^({IPV4_REG}|{IPV6_REG})" r"/\d{1,3}(?![\n\r])$"),
        description="An IPv4 or IPv6 network prefix.",
    ),
]

MACAddress = Annotated[
    str | None,
    Field(
        pattern=re.compile(r"^[a-fA-F\d]{2}(:[a-fA-F\d]{2}){5}(?![\n\r])$"),
        description="MAC address in Linux format.",
        examples=["00:11:22:33:44:55"],
    ),
]

Hostname = Annotated[
    str,
    Field(
        pattern=re.compile(HOST_REG),
        description="A Linux hostname (not FQDN).",
    ),
]

LinuxInterfaceName = Annotated[
    str,
    Field(
        pattern=re.compile(rf"^{LINT_REG}(?![\n\r])$"),
        description="Interface name or number in a Linux host.",
    ),
]


class OldPermission(StrEnum):
    READ_ONLY = DefaultPermissions.READ_ONLY
    READ_WRITE = DefaultPermissions.READ_WRITE


class Permission(StrEnum):
    LAB_ADMIN = DefaultPermissions.LAB_ADMIN
    LAB_EDIT = DefaultPermissions.LAB_EDIT
    LAB_EXEC = DefaultPermissions.LAB_EXEC
    LAB_VIEW = DefaultPermissions.LAB_VIEW


Permissions = Annotated[
    list[Permission],
    Field(
        ...,
        description="Permission array.",
        examples=[[Permission.LAB_ADMIN, Permission.LAB_EXEC]],
    ),
    AfterValidator(_remove_duplicates),
]

EffectivePermissions = Annotated[
    Permissions,
    Field(
        description="Effective permissions for the current user.",
    ),
]


class BorderStyle(StrEnum):
    SOLID = ""
    DOTTED = "2,2"
    DASHED = "4,2"


COLOR_EXAMPLES = ["#FF00FF", "rgba(255, 0, 0, 0.5)", "lightgoldenrodyellow"]
COLOR_EXAMPLES_STR = f"(e.g.,`{'` or `'.join(COLOR_EXAMPLES)}`)"

AnnotationColor = Annotated[
    str, Field(min_length=0, max_length=32, examples=COLOR_EXAMPLES)
]


DateTimeString = Annotated[
    datetime,
    Field(
        description="Datetime string in ISO8601 format.",
        examples=["2021-02-28T07:33:47+00:00"],
    ),
]

Label = Annotated[
    str,
    Field(
        min_length=1,
        max_length=128,
        description="A label.",
        examples=["Any human-readable text"],
    ),
]


DefinitionID = Annotated[
    str,
    Field(
        min_length=1,
        max_length=250,
        pattern=re.compile(r"^(?![.])[^!@#%^&*();$\n\r\t/\\]{1,250}(?![\n\r])$"),
        description="Name of the node or image definition (max 250 UTF-8 bytes).",
        examples=["server"],
    ),
]

Tag = Annotated[str, Field(max_length=64, description="A tag.")]

TagArray = Annotated[list[Tag], Field(description="Array of string tags.")]

Coordinate = Annotated[int, Field(description="A coordinate.", ge=-15000, le=15000)]

PinnedComputeID = Annotated[
    UUID4Type | None,
    Field(
        description="The ID of the compute host where this node is to be exclusively deployed."
    ),
]


class States(StrEnum):
    DEFINED_ON_CORE = "DEFINED_ON_CORE"
    STOPPED = "STOPPED"
    STARTED = "STARTED"


State = Annotated[States, Field(description="The state of the element.")]


# should be in sync with BaseDbMixin from simple core
class BaseDBModel(BaseModel):
    """Base class for all database based models."""

    # TODO: remove default=None from all of these
    id: UUID4Type = Field(default=None)
    created: DateTimeString = Field(
        default=None,
        description="The create date of the object, a string in ISO8601 format.",
    )
    modified: DateTimeString = Field(
        default=None,
        description="Last modification date of the object, a string in ISO8601 format.",
    )

    @field_serializer("created", "modified")
    def serialize_datetime(self, field: datetime) -> str:
        return field.astimezone(UTC).isoformat(timespec="seconds")


class FreeFormSchema(BaseModel, extra="allow"):
    """A freeform JSON object."""

    pass



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#

import os
from pathlib import Path

MIME_JSON = "application/json"
MIME_TEXT = "text/plain"
MIME_YAML = "application/x-yaml"

DEFAULT_VERSION = "unknown"
DEFAULT_PYATS_USERNAME = "cisco"
DEFAULT_PYATS_PASSWORD = "cisco"

BYTES_IN_MB = 1024**2
BYTES_IN_GB = 1024**3

EVENT_LIMIT = 10000
EVENT_CLEAR = int(EVENT_LIMIT * 0.05)

# from /usr/include/sys/syslog.h
# #define LOG_LOCAL6      (22<<3) /* reserved for local use */
# Python defines it like C, SyslogHandler shifts it and adds priority
TELEMETRY_LOG_LEVEL = 22

SSHD_CLUSTER_PORT = 1222

BASE_DIR = Path(os.environ.get("BASE_DIR", "/var/local/virl2"))
KNOWN_HOSTS = BASE_DIR / "secrets/admin_hosts/known_hosts"
COCKPIT_MACHINES = Path("/etc/cockpit/machines.d")
DROPFOLDER_DIRECTORY = os.environ.get("DROPFOLDER_DIRECTORY", str(BASE_DIR / "dropfolder"))
LIBVIRT_IMAGES = os.environ.get("LIBVIRT_IMAGES", "/var/lib/libvirt/images")
IMAGE_DEFINITION_DIRECTORY = f"{LIBVIRT_IMAGES}/virl-base-images"
NODE_DEFINITION_DIRECTORY = f"{LIBVIRT_IMAGES}/node-definitions"
USER_IMAGES = os.environ.get("USER_IMAGES", str(BASE_DIR / "images"))
DEFAULT_VIRL2 = "/etc/default/virl2"
IMAGES_STATE = BASE_DIR / "base_images.state"
# name of the node configuration customizer script
CONFIG_CUSTOMIZER = "cml-customizer.sh"

CML_NAMESPACE_URI = "http://cisco.com/cml"
CML_NS_PREFIX = "cml"

# Prefix used for unmanaged switch interfaces
UMS_PREFIX = "ums-"

# disk image formats supported for VMs
QCOW2 = "qcow2"
QCOW = "qcow"
IOL = "iol"
TAR = "tar"
TARGZ = "tar.gz"
SUPPORTED_IMAGE_FORMATS = [QCOW2, QCOW, IOL, TAR, TARGZ]



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#

import re
from typing import Annotated

from fastapi import Path
from pydantic import BaseModel, Field, model_validator

from cml_mcp.schemas.simple_common.constants import BASE_DIR
from cml_mcp.schemas.common import HTTP_URL_REG, DefinitionID, UUID4Type
from cml_mcp.schemas.labs import LabDescription, LabTitle

GitUrl = Annotated[str, Field(pattern=re.compile(HTTP_URL_REG), description="An HTTPS git URL")]

RepoName = Annotated[
    str,
    Field(
        description="The name of the repository.",
        examples=["cml-labs"],
        min_length=1,
        max_length=64,
        pattern=re.compile(r"^(?![/.])[\w.-]{1,64}(?![\n\r])$"),
    ),
]

RepoFolder = Annotated[
    str,
    Field(
        description="The name of the folder to clone in the repository.",
        examples=["cml-community"],
        min_length=1,
        max_length=255,
        pattern=re.compile(r"^(?![/.])[\w./-]{1,255}(?![\n\r])$"),
    ),
]

RepoIdPathParameter = Annotated[UUID4Type, Path(description="The ID of an available lab repository.")]

SampleLabIdPathParameter = Annotated[UUID4Type, Path(description="The ID of the specified sample lab.")]


class CreateLabRepo(BaseModel, extra="forbid"):
    """Lab repository creation"""

    url: GitUrl = Field(..., description="The URL of the repository.")
    name: RepoName = Field(..., description="The name of repository.")
    folder: RepoFolder | None = Field(
        default=None,
        description="Limit the git pull to a single folder in the repository.",
    )

    @model_validator(mode="after")
    def repo_path_validation(self):
        root = (BASE_DIR / "sample-labs").resolve()
        repo = (root / self.name).resolve()
        if not repo.is_relative_to(root):
            raise ValueError("Name is not relative the repository root.")
        if self.folder and not (repo / self.folder).resolve().is_relative_to(repo):
            raise ValueError("Folder is not relative to the repository.")
        return self


class LabRepoResponse(CreateLabRepo, extra="forbid"):
    id: UUID4Type = Field(..., description="ID of the lab repository.")


class SampleLabResponse(BaseModel, extra="forbid"):
    id: UUID4Type
    title: LabTitle
    description: LabDescription
    name: RepoName
    node_types: list[DefinitionID]
    file_path: str = Field(..., description="The relative path of the sample lab.")


class LabReposRefreshStatus(BaseModel, extra="forbid"):
    success: bool = Field(..., description="The status of the repository refresh.")
    message: str = Field(..., description="Description of the status of the repository refresh.")
    index: str = Field(
        default="Index file updated successfully",
        description="The index file status of the repository.",
    )



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from datetime import datetime
from enum import StrEnum
from typing import Annotated

from fastapi import Body, Path
from pydantic import BaseModel, Field, model_validator

from cml_mcp.schemas.common import IPAddress, MACAddress, UUID4Type

LINK_ENCAP_DESCRIPTION = "Link encapsulation"

CaptureKeyPathParameter = Annotated[UUID4Type, Path(description="The UUID of the PCAP (link capture key).")]


# maximum: 1000000 # this is random
PacketIdPathParameter = Annotated[
    int,
    Path(
        description="The numeric ID of a specific packet within the PCAP.",
        ge=1,
        le=1000000,
        examples=[4712],
    ),
]


# The enum values presented here need to be in sync with the UI in
# LinkPacketCapture.vue and the fabric in link type.
class LinkEncaps(StrEnum):
    ethernet = "ethernet"
    frelay = "frelay"
    ppp = "ppp"
    ppp_hdlc = "ppp_hdlc"
    pppoe = "pppoe"
    c_hdlc = "c_hdlc"
    slip = "slip"
    ax25 = "ax25"


class PCAPStart(BaseModel, extra="forbid"):
    maxpackets: int = Field(
        default=None,
        description="Maximum amount of packets to be captured.",
        ge=1,
        le=1000000,
        examples=[50],
    )
    maxtime: int = Field(
        default=None,
        description="Maximum time (seconds) the PCAP can run.",
        ge=1,
        le=86400,
        examples=[60],
    )
    bpfilter: str = Field(
        default=None,
        description="Berkeley packet filter.",
        min_length=1,
        max_length=128,
        examples=["src 0.0.0.0"],
    )
    encap: LinkEncaps = Field(default=None, description=LINK_ENCAP_DESCRIPTION)

    @model_validator(mode="after")
    def check_at_least_one(self):
        if self.maxpackets is None and self.maxtime is None:
            raise ValueError("Either 'maxpackets' or 'maxtime' must be specified")
        return self


PCAPStartBody = Annotated[PCAPStart, Body(description="Send parameters in JSON format to start PCAP.")]


class PCAPBaseConfigStatus(BaseModel, extra="forbid"):
    maxpackets: int = Field(
        default=None,
        description="Maximum amount of packets to be captured.",
        ge=0,
        le=1000000,
        examples=[50],
    )
    maxtime: int | None = Field(
        default=None,
        description="Maximum time (seconds) the PCAP can run.",
        ge=0,
        le=86400,
        examples=[60],
    )
    bpfilter: str | None = Field(
        default=None,
        description="Berkeley packet filter.",
        min_length=1,
        max_length=128,
        examples=["src 0.0.0.0"],
    )
    encap: LinkEncaps | None = Field(default=None, description=LINK_ENCAP_DESCRIPTION)


class PCAPConfigStatus(PCAPBaseConfigStatus, extra="forbid"):
    link_capture_key: UUID4Type | None = Field(
        ...,
        description="Key or ID for the packet capture running on the specified link.",
    )


class PCAPStatusResponse(BaseModel, extra="forbid"):
    config: PCAPConfigStatus | None = Field(
        None,
        description="The configuration of the PCAP. Empty when PCAP is not running",
    )
    starttime: datetime | None = Field(None, description="The start time of the PCAP. None when PCAP is not running")
    packetscaptured: int | None = Field(
        None,
        description="The number of packets captured. None when PCAP is not running",
        ge=0,
    )
    encap: LinkEncaps | None = Field(default=None, description=LINK_ENCAP_DESCRIPTION)


class PCAPItem(BaseModel, extra="forbid"):
    no: str = Field(
        ...,
        description="Packet number",
        min_length=1,
        max_length=16,
        examples=["1", "2", "3"],
    )
    time: str = Field(
        ...,
        description="Time since PCAP was started",
        min_length=1,
        max_length=16,
        examples=["12.003743"],
    )
    source: MACAddress | IPAddress = Field(..., description="The MAC/IP address of the source.")
    destination: MACAddress | IPAddress = Field(..., description="The MAC/IP address of the destination.")
    length: str = Field(
        ...,
        description="The length of the packet.",
        min_length=1,
        max_length=16,
        examples=["64"],
    )
    protocol: str = Field(
        ...,
        description="Protocol of the packet.",
        min_length=1,
        max_length=16,
    )
    info: str = Field(
        ...,
        description="Information about the packet.",
        min_length=1,
        max_length=512,
    )



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from typing import Annotated

from pydantic import BaseModel, Field

from cml_mcp.schemas.common import UUID4Type


class Error(BaseModel):
    code: int = Field(..., description="The HTTP status that was associated with this error.")
    description: str = Field(..., description="A human-readable message that describes the error.")


class ErrorResponse(Error):
    pass


class FreeFormResponse(BaseModel, extra="allow"):
    """A freeform JSON object."""

    pass


class IdObject(BaseModel):
    id: UUID4Type = Field(...)


class IdResponse(IdObject, extra="forbid"):
    """Successful POST / Create operation.
    The response body is a JSON object that indicates the ID of the created object."""

    pass


ConvergenceResponse = Annotated[
    bool,
    Field(description="The response body is a JSON object with a boolean indicating if convergence has occurred"),
]
